from terraform_compliance.common.helper import (
    check_if_cidr,
    is_ip_in_cidr
)
from terraform_compliance.common.exceptions import (
    TerraformComplianceInvalidData,
    TerraformComplianceInternalFailure
)


class SecurityGroupRule(object):
    def __init__(self, **kwargs):
        self.from_port = kwargs.get('from_port', kwargs.get('values', {}).get('from_port'))
        self.to_port = kwargs.get('to_port', kwargs.get('values', {}).get('to_port'))
        self.ports = str(kwargs.get('port'))
        self.protocol = [proto for proto in [kwargs.get('protocol', kwargs.get('values', {}).get('protocol', '-1'))]]
        self.cidr_blocks = kwargs.get('cidr_blocks', kwargs.get('values', {}).get('cidr_blocks', []))
        self.address = kwargs.get('values', {}).get('address')
        self.description = kwargs.get('description', kwargs.get('values', {}). get('description'))
        self.port_is_single = False

        self._normalise_protocols()
        self._normalise_ports()
        self._check_for_invalidations()

    def _normalise_ports(self):
        # Sometimes terraform reports ports like this
        if self.from_port == 0 and self.to_port == 0:
            self.to_port = 65535
            self.port_is_single = False

        if not self.from_port and not self.to_port:

            # There is no port information, so we need to set everything to ALL, because
            # terraform does that automatically.
            if self.ports == 'None':
                self.from_port = 1
                self.to_port = 65535
            else:
                self._expand_ports(self.ports)

        else:
            self.from_port = int(self.from_port)
            self.to_port = int(self.to_port)

            self.from_port = self.from_port if self.from_port > 0 else 1
            self.to_port = self.to_port if self.to_port > 0 else 1
            self.ports = self._get_port_range(self.from_port, self.to_port)

            self.port_is_single = True if self.from_port == self.to_port else False

    def _normalise_protocols(self):
        # TODO: Implement other protocols, not just TCP and UDP.
        # There is always one protocol in the list as an input.
        if self.protocol[0] == '-1' or isinstance(self.protocol[0], int):
            self.protocol = ['tcp', 'udp']
        else:
            self.protocol[0] = self.protocol[0].lower()

    def _check_for_invalidations(self):
        if (self.from_port and self.to_port) and self.from_port > self.to_port:
            raise TerraformComplianceInvalidData('Invalid configuration from_port can not be bigger than to_port. '
                                                 '{} > {} {} in {}'.format(self.from_port,
                                                                           self.to_port,
                                                                           self.protocol,
                                                                           self.cidr_blocks))

        if self.cidr_blocks is None:
            raise TerraformComplianceInvalidData('A cidr range must be given.')

        if isinstance(self.cidr_blocks, list):
            for cidr in self.cidr_blocks:
                if check_if_cidr(cidr) is False:
                    raise TerraformComplianceInvalidData('{} is not a valid CIDR.'.format(cidr))
        elif isinstance(self.cidr_blocks, str):
            if check_if_cidr(self.cidr_blocks) is False:
                raise TerraformComplianceInvalidData('{} is not a valid CIDR.'.format(self.cidr_blocks))
        else:
            raise TerraformComplianceInvalidData('Invalid CIDR Type {} : {}'.format(type(self.cidr_blocks),
                                                                                    self.cidr_blocks))

        if not self.protocol:
            raise TerraformComplianceInvalidData('A protocol must be given.')

    def _expand_ports(self, port):
        if '-' in port:
            self.from_port, self.to_port = [int(port_number) for port_number in port.split('-')]
            self.from_port = self.from_port if self.from_port > 0 else 1
            self.to_port = self.to_port if self.to_port > 0 else 1
            self.ports = self._get_port_range(self.from_port, self.to_port)
            self.port_is_single = False

        elif ',' in port:
            self.from_port = None
            self.to_port = None
            self.port_is_single = False
            self.ports = set([int(port_number) for port_number in port.split(',') if int(port_number) > 0])

        elif port.isnumeric():
            self.from_port = self.to_port = int(port)
            self.ports = self._get_port_range(self.from_port, self.to_port)
            self.port_is_single = True

        else:
            raise TerraformComplianceInternalFailure('Unexpected situation. '
                                                     'Please report this port structure: {}'.format(vars(self)))

    @staticmethod
    def _get_port_range(from_port, to_port):
        return set(range(from_port, to_port+1))


class SecurityGroup(object):
    def __init__(self, given_reqs, security_groups_in_plan, address='test_sg'):
        self.given = given_reqs
        self.sgs = self._clean(security_groups_in_plan)
        self.sgs = [self.sgs] if isinstance(self.sgs, dict) else self.sgs
        self.address = address

        self.exact_match = False
        self.negative_match = True
        self.include_match = False
        self.singular_check = True
        self.multiple_check = False

        self.given_rule = SecurityGroupRule(**given_reqs)
        self.plan_rules = [SecurityGroupRule(**rule_data) for rule_data in self.sgs if rule_data.get('cidr_blocks')]

        self.found_ports = set()

    def must_only_have(self):
        self.exact_match = True
        self.negative_match = False
        self.include_match = False
        self.singular_check = False

    # This is the default for validate() method
    def must_not_have(self):
        self.exact_match = False
        self.negative_match = True
        self.include_match = False
        self.singular_check = True

    def must_have(self):
        self.exact_match = False
        self.negative_match = False
        self.include_match = True
        self.singular_check = False

    def validate(self):

        #
        # Make it flexible enough to support other scenarios
        # Currently we have ;
        #
        # 1. "must only" ports, which means all ports defined in ALL security group rules must have exactly the same
        #    ports.
        # 2. "must not" ports, which means the ports defined in ANY security group, should not have this port with cidr
        #    , etc.
        # 3. "must have" ports, which means, the ports defined in ALL security groups should include these ports. < This
        #    doesn't exist yet.
        #

        ports_found = set()
        for rule in self.plan_rules:
            # General Controls that is applicable for every scenario
            network_check = self._validate_network(rule.cidr_blocks)
            ports_check = self._validate_ports(rule)
            protocol_check = self._validate_protocol(rule)

            # Apply enforcement on every rule. Only applicable for 'must not have'
            if self.singular_check:
                result, error = self._run_singular_validation(rule, network_check, ports_check, protocol_check)
                if result is False:
                    return result, error

            # Apply enforcement after the for loop, gather information first.
            elif not self.singular_check and network_check and protocol_check:
                # Gather all ports defined within the Security Group rule that matches CIDR.
                ports_found = ports_found | rule.ports

        # Finalise singular checks as returning True if nothing happens,
        # since it must have failed already.
        if self.singular_check:
            return True, None

        # Process non singular checks
        else:
            # So if the scenario is 'must only have'
            # Then sets must be exactly equal to each other.
            # If not, first get GIVEN->FOUND, FOUND->GIVEN and no disjoint differences
            # then return an error output
            errors = []
            if self.exact_match:
                if ports_found == self.given_rule.ports:
                    return True, None

                # Create errors
                given_found_diff = self.given_rule.ports - ports_found
                found_given_diff = ports_found - self.given_rule.ports
                nodisjoint = self.given_rule.ports.isdisjoint(ports_found)

                if given_found_diff:
                    e = self._prepare_output(given_found_diff)
                    errors.append('{}/{} port{} not defined within {} network{} in {}.'.format(e['protocol'],
                                                                                               e['ports'],
                                                                                               e['ports_plural'],
                                                                                               e['cidr_blocks'],
                                                                                               e['cidr_blocks_plural'],
                                                                                               e['address']))

                if found_given_diff:
                    e = self._prepare_output(found_given_diff)
                    errors.append('{}/{} port{} defined within {} network{} in {}.'.format(e['protocol'],
                                                                                           e['ports'],
                                                                                           e['ports_plural'],
                                                                                           e['cidr_blocks'],
                                                                                           e['cidr_blocks_plural'],
                                                                                           e['address']))

                if nodisjoint:
                    e = self._prepare_output(set())
                    errors.append('None of the ports given defined within {} network{} in '
                                  '{}.'.format(e['cidr_blocks'],
                                               e['cidr_blocks_plural'],
                                               e['address']))

            # If the scenario is 'must have', then given ports must be a subset (or equal) of ports_found
            # If not, found which ports are missing via GIVEN->FOUND diff
            else:
                if self.given_rule.ports <= ports_found:
                    return True, None

                given_found_diff = self.given_rule.ports - ports_found

                if given_found_diff:
                    e = self._prepare_output(given_found_diff)
                    errors.append('{}/{} port{} not defined within {} network{} in {}.'.format(e['protocol'],
                                                                                               e['ports'],
                                                                                               e['ports_plural'],
                                                                                               e['cidr_blocks'],
                                                                                               e['cidr_blocks_plural'],
                                                                                               e['address']))
            if errors:
                return False, '{}'.format("\n".join(errors))

    def _run_singular_validation(self, rule, network_check, ports_check, protocol_check):
        if self.negative_match and network_check and ports_check and protocol_check:
            e = self._prepare_output(rule, ports_check, protocol_check)
            return False, '{}/{} port{} defined within {} network{} in {}.'.format(e['protocol'],
                                                                                   e['ports'],
                                                                                   e['ports_plural'],
                                                                                   e['cidr_blocks'],
                                                                                   e['cidr_blocks_plural'],
                                                                                   e['address'])
        else:
            return True, None

    def _validate_network(self, cidr_in_plan):
        return is_ip_in_cidr(self.given_rule.cidr_blocks, cidr_in_plan)

    def _validate_ports(self, rule_data):
        return self.given_rule.ports & rule_data.ports

    def _validate_protocol(self, rule_data):
        return set(self.given_rule.protocol) & set(rule_data.protocol)

    def _prepare_output(self, rule, ports_check=None, protocol_check=None):
        output = dict()

        if self.singular_check:
            output['protocol'] = str(list(protocol_check)[0]) if len(protocol_check) == 1 else '({})'.format(','.join(str(s) for s in protocol_check))
            output['cidr_blocks'] = rule.cidr_blocks[0] if len(rule.cidr_blocks) == 1 else ', '.join(rule.cidr_blocks)
            output['cidr_blocks_plural'] = '' if len(rule.cidr_blocks) == 1 else 's'
            output['address'] = rule.address if rule.address else self.address
            output['ports'] = str(list(ports_check)[0]) if len(ports_check) == 1 else '({})'.format(','.join(str(s) for s in ports_check))
            # TODO: If the ports output is too large, make it smaller by showing a range only.
            output['ports_plural'] = ' is' if len(ports_check) == 1 else 's are'
        else:
            output['protocol'] = str(list(self.given_rule.protocol)[0]) if len(self.given_rule.protocol) == 1 else '({})'.format(','.join(str(s) for s in self.given_rule.protocol))
            output['cidr_blocks'] = self.given_rule.cidr_blocks
            output['cidr_blocks_plural'] = ''
            output['address'] = self.address
            output['ports'] = str(list(rule)[0]) if len(rule) == 1 else '({})'.format(','.join(str(s) for s in rule))
            # TODO: If the ports output is too large, make it smaller by showing a range only.
            output['ports_plural'] = ' is' if len(rule) == 1 else 's are'

        return output

    def _clean(self, sg_array):
        if isinstance(sg_array, list):
            for sg in sg_array:
                self._clean(sg)

        elif isinstance(sg_array, dict):
            if 'self' in sg_array:
                sg_array.pop('self')

        return sg_array



from terraform_compliance.common.defaults import Defaults


def look_for_bdd_tags(_step_obj):
    _step_obj.context.no_failure = False
    _step_obj.context.failure_class = None

    if hasattr(_step_obj, 'all_tags') and len(_step_obj.all_tags) > 0:
        for tag in _step_obj.all_tags:
            if tag.name.lower() in Defaults().no_failure_tags:
                _step_obj.context.no_failure = True
                _step_obj.context.failure_class = tag.name

    return _step_obj



from unittest import TestCase
from terraform_compliance.common.readable_dir import ReadableDir
import os
from mock import patch


class TestReadableDir(TestCase):
    def test_readable_dir_call_git_repository_success(self):
        class Namespace(object):
            pass

        resp = ReadableDir('parser', 'value').__call__('parser', Namespace, 'git:value')
        self.assertEqual(Namespace.value, 'value')
        self.assertTrue(resp)


    @patch.object(os.path, 'isdir', return_value=False)
    def test_readable_dir_call_not_a_local_directory(self, *args):
        class Namespace(object):
            pass

        with self.assertRaises(SystemExit):
            ReadableDir('parser', 'value').__call__('parser', Namespace, 'non_existing_dir')

    @patch.object(os.path, 'isdir', return_value=True)
    @patch.object(os, 'access', return_value=True)
    def test_readable_dir_call_is_a_local_dir(self, *args):
        class Namespace(object):
            pass

        resp = ReadableDir('parser', 'value').__call__('parser', Namespace, 'value')
        self.assertEqual(Namespace.value, 'value')
        self.assertTrue(resp)

    @patch.object(os.path, 'isdir', return_value=True)
    @patch.object(os, 'access', return_value=False)
    def test_readable_dir_call_can_not_access_dir(self, *args):
        class Namespace(object):
            pass

        with self.assertRaises(SystemExit):
            ReadableDir('parser', 'value').__call__('parser', Namespace, 'non_accessible_dir')




import json
from terraform_compliance.common.helper import seek_key_in_dict, flatten_list
import sys
from copy import deepcopy


class TerraformParser(object):
    def __init__(self, filename, parse_it=True):
        '''
        This class reads the given terraform plan filename ( in json format ) and assigns required variables for
        further steps in terraform-compliance. If the file is not a json or terraform plan file, then it will be
        checked and exited in prior steps.

        :param filename: terraform plan filename in json format.
        :parse_it: Runs self.parse() if given.

        :return: None
        '''
        self.supported_terraform_versions = (
            '0.12', # This is here because this tuple must have multiple values.
            '0.12.'
        )
        self.supported_format_versions = ['0.1']

        self.raw = self._read_file(filename)
        self.variables = None
        self.resources = {}

        self.data = {}

        self.providers = {}

        self.configuration = dict(resources={}
                                  , variables={}
                                  )
        self.file_type = "plan"
        self.resources_raw = {}

        if parse_it:
            self.parse()

    def _version_check(self):
        if self.raw['format_version'] not in self.supported_format_versions:
            print('\nFATAL ERROR: Unsupported terraform plan output format version '
                  '({}).\n'.format(self.raw['format_version']))
            sys.exit(1)

        if not self.raw['terraform_version'].startswith(self.supported_terraform_versions):
            print('\nFATAL ERROR: Unsupported terraform version '
                  '({}).\n'.format(self.raw['terraform_version']))
            sys.exit(1)

        return True

    def _identify_data_file(self):
        if 'values' in self.raw:
            self.file_type = 'state'

    def _read_file(self, filename):
        '''
        Reads the json filename as a dictionary. We are not checking if the file is a json file again, since
        it is already checked in main.py

        :param filename: json filename with full path
        :return: parsed dictionary
        '''
        with open(filename, 'r', encoding='utf-8') as plan_file:
            data = json.load(plan_file)

        return data

    def _parse_variables(self):
        '''
        Assignes all variables that is defined within the terraform plan

        :return: none
        '''
        self.variables = self.raw.get('variables', {})

    def _parse_resources(self):
        '''
        Assigns all resources defined in the terraform plan

        :return: none
        '''

        # Resources ( exists in Plan )
        for findings in seek_key_in_dict(self.raw.get('planned_values', {}).get('root_module', {}), 'resources'):
            for resource in findings.get('resources', []):
                if resource['address'].startswith('data'):
                    self.data[resource['address']] = resource
                else:
                    self.resources[resource['address']] = resource

        # Resources ( exists in State )
        for findings in seek_key_in_dict(self.raw.get('values', {}).get('root_module', {}), 'resources'):
            for resource in findings.get('resources', []):
                if resource['address'].startswith('data'):
                    self.data[resource['address']] = resource
                else:
                    self.resources[resource['address']] = resource

        # Resources ( exists in Prior State )
        for findings in seek_key_in_dict(self.raw.get('prior_state', {}).get('values', {}).get('root_module', {}).get('resources', {}), 'resource'):
            for resource in findings.get('resources', []):
                if resource['address'].startswith('data'):
                    self.data[resource['address']] = resource
                else:
                    self.resources[resource['address']] = resource

        # Child Modules Resources ( exists in State )
        for findings in seek_key_in_dict(self.raw.get('values', {}).get('root_module', {}), 'child_modules'):
            for resource in findings.get('resources', []):
                if resource['address'].startswith('data'):
                    self.data[resource['address']] = resource
                else:
                    self.resources[resource['address']] = resource

        # Resource Changes ( exists in Plan )
        for finding in self.raw.get('resource_changes', {}):
            resource = deepcopy(finding)
            change = resource.get('change', {})
            actions = change.get('actions', [])
            if actions != ['delete']:
                resource['values'] = change.get('after', {})
                if 'change' in resource:
                    del resource['change']

                if resource['address'].startswith('data'):
                    self.data[resource['address']] = resource
                else:
                    self.resources[resource['address']] = resource

    def _parse_configurations(self):
        '''
        Assigns all configuration related data defined in the terraform plan. This is mostly used for
        resources referencing each other.

        :return: none
        '''

        # Resources
        self.configuration['resources'] = {}

        for findings in seek_key_in_dict(self.raw.get('configuration', {}).get('root_module', {}), 'resources'):
            for resource in findings.get('resources', []):
                if resource['address'].startswith('data'):
                    self.data[resource['address']] = resource
                else:
                    self.configuration['resources'][resource['address']] = resource

        # Variables
        self.configuration['variables'] = {}
        for findings in seek_key_in_dict(self.raw.get('configuration', {}).get('root_module', {}), 'variables'):
            self.configuration['variables'] = findings.get('variables')

        # Providers
        self.configuration['providers'] = {}
        for findings in seek_key_in_dict(self.raw.get('configuration', {}), 'provider_config'):
            self.configuration['providers'] = findings.get('provider_config', {})

        # Outputs
        self.configuration['outputs'] = {}
        for findings in seek_key_in_dict(self.raw.get('configuration', {}), 'outputs'):
            for key, value in findings.get('outputs', {}).items():
                tmp_output = dict(address=key, value={})
                if 'expression' in value:
                    if 'references' in value['expression']:
                        tmp_output['value'] = value['expression']['references']
                        tmp_output['type'] = 'object'
                    elif 'constant_value' in value['expression']:
                        tmp_output['value'] = value['expression']['constant_value']

                if 'sensitive' in value:
                    tmp_output['sensitive'] = str(value['sensitive']).lower()
                else:
                    tmp_output['sensitive'] = 'false'

                if 'type' in value:
                    tmp_output['type'] = value['type']
                elif 'type' not in tmp_output:
                    if isinstance(tmp_output['value'], list):
                        tmp_output['type'] = 'list'
                    elif isinstance(tmp_output['value'], dict):
                        tmp_output['type'] = 'map'
                    elif isinstance(tmp_output['value'], str):
                        tmp_output['type'] = 'string'
                    elif isinstance(tmp_output['value'], int):
                        tmp_output['type'] = 'integer'
                    elif isinstance(tmp_output['value'], bool):
                        tmp_output['type'] = 'boolean'

                self.configuration['outputs'][key] = tmp_output

    def _mount_resources(self, source, target, ref_type):
        '''
        Mounts values of the source resource to the target resource's values with ref_type key

        :param source: source resource
        :param target:  target resource
        :param ref_type: reference type (e.g. ingress )
        :return: none
        '''
        for source_resource in source:

            if 'values' not in self.resources[source_resource]:
                continue

            for target_resource in target:

                if target_resource not in self.resources or 'values' not in self.resources[target_resource]:
                    continue

                resource = deepcopy(self.resources[source_resource]['values'])
                resource['terraform-compliance.mounted'] = True

                if 'terraform-compliance.mounted_resources' not in self.resources[target_resource]:
                    self.resources[target_resource]['terraform-compliance.mounted_resources'] = []

                if 'terraform-compliance.mounted_resources.addresses' not in self.resources[target_resource]:
                    self.resources[target_resource]['terraform-compliance.mounted_resources.addresses'] = []

                if ref_type not in self.resources[target_resource]['values']:
                    self.resources[target_resource]['values'][ref_type] = []
                    self.resources[target_resource]['values'][ref_type].append(resource)
                    self.resources[target_resource]['terraform-compliance.mounted_resources'].append(ref_type)
                    self.resources[target_resource]['terraform-compliance.mounted_resources.addresses'].extend(source)
                else:
                    self.resources[target_resource]['values'][ref_type].append(resource)
                    self.resources[target_resource]['terraform-compliance.mounted_resources'].append(ref_type)
                    self.resources[target_resource]['terraform-compliance.mounted_resources.addresses'].extend(source)

    def _find_resource_from_name(self, resource_name):
        '''
        Finds all the resources that is starting with resource_name

        :param resource_name: The first initials of the resource
        :return: list of the found resources
        '''
        if resource_name in self.resources:
            return [resource_name]

        resource_list = []

        resource_type, resource_id = resource_name.split('.')[0:2]

        if resource_type == 'module':
            module_name, output_id = resource_name.split('.')[1:3]
            module = self.raw['configuration']['root_module'].get('module_calls', {}).get(module_name, {})

            output_value = module.get('module', {}).get('outputs', {}).get(output_id, {})
            resources = output_value.get('expression', {}).get('references') if 'expression' in output_value else output_value.get('value', [])
            resources = ['{}.{}.{}'.format(resource_type, module_name, res) for res in resources]

            if resources:
                resource_list.extend(resources)
        else:
            for key, value in self.resources.items():
                if value['type'] == resource_type and value['name'] == resource_id:
                    resource_list.append(key)

        return resource_list

    def _mount_references(self):
        '''
        Find the references that is defined in self.configuration
        :return:
        '''
        self.resources_raw = deepcopy(self.resources)
        invalid_references = ('var')

        # This section will link resources found in configuration part of the plan output.
        # The reference should be on both ways (A->B, B->A) since terraform sometimes report these references
        # in opposite ways, depending on the provider structure.
        for resource in self.configuration['resources']:
            if 'expressions' in self.configuration['resources'][resource]:
                ref_list = []
                for key, value in self.configuration['resources'][resource]['expressions'].items():
                    if 'references' in value:
                        ref_list.extend([self._find_resource_from_name(ref) for ref in value['references']
                                         if not ref.startswith(invalid_references)])

                if ref_list:
                    ref_type = self.configuration['resources'][resource]['expressions'].get('type',
                                                                                            {}).get('constant_value',
                                                                                                    {})

                    if not ref_type and not resource.startswith(('data')):
                            resource_type, resource_id = resource.split('.')
                            ref_type = resource_type

                    # Mounting A->B
                    source_resources = self._find_resource_from_name(self.configuration['resources'][resource]['address'])
                    self._mount_resources(source_resources, flatten_list(ref_list), ref_type)
                    
                    # Mounting B->A
                    for source_resource in flatten_list(ref_list):
                        if not source_resource.startswith(('var', 'data', 'module', 'provider')):
                            ref_type = source_resource.split('.', maxsplit=1)[0]

                            self._mount_resources([source_resource],
                                                  source_resources,
                                                  ref_type)

    def _distribute_providers(self):
        for resource_name, resource_data in self.resources.items():
            resource_provider = resource_name.split('_')[0]

            if resource_provider not in self.providers:
                self.providers[resource_provider] = {}


            self.providers[resource_provider][resource_name] = resource_data

    def parse(self):
        '''
        Main method for initialising the parsing of the terraform plan json file

        :return: nothing
        '''
        self._version_check()
        self._identify_data_file()
        self._parse_resources()

        if self.file_type == 'plan':
            self._parse_variables()
            self._parse_configurations()

        self._mount_references()
        self._distribute_providers()
        return

    def find_resources_by_type(self, resource_type):
        '''
        Finds all resources matching with the resource_type

        :param resource_type: String of resource type defined in terraform
        :return: list of dict including resources
        '''
        resource_list = []

        for _, resource_data in self.resources.items():
            if resource_type == 'any' or (resource_data['type'] == resource_type.lower() and resource_data['mode'] == 'managed'):
                resource_list.append(resource_data)

        return resource_list

    def find_data_by_type(self, resource_type):
        '''
        Finds all data matching with the resource_type

        :param resource_type: String of resource type defined in terraform
        :return: list of dict including resources
        '''
        resource_list = []

        for _, resource_data in self.data.items():
            if resource_data['type'] == resource_type.lower():
                resource_list.append(resource_data)

        return resource_list

    def get_providers_from_configuration(self, provider_type):
        '''
        Returns all providers as a list for the given provider type

        :param provider_type: String of a provider type like aws
        :return: list of providers that has this type
        '''
        providers = []
        for provider_alias, values in self.configuration['providers'].items():
            if isinstance(values, dict) and values.get('name') == provider_type:
                providers.append(values)

        return providers



from unittest import TestCase
from terraform_compliance.common.pip_helper import subprocess, reinstall_radish
from mock import patch


class TestPipHelper(TestCase):

    @patch.object(subprocess, 'call', return_value=None)
    def test_reinstall_radish(self, *args):
        with self.assertRaises(SystemExit) as context:
            reinstall_radish()



from unittest import TestCase
from terraform_compliance.steps.steps import (
    i_have_name_section_configured,
    i_action_them,
    i_expect_the_result_is_operator_than_number,
    it_condition_contain_something,
    property_is_enabled,
    its_value_condition_match_the_search_regex_regex,
    it_condition_have_proto_protocol_and_port_port_for_cidr,
    it_fails,
    its_key_is_value, its_key_is_not_value,
    its_value_condition_equal,
    its_value_condition_contain
)
from terraform_compliance.common.exceptions import TerraformComplianceNotImplemented, Failure
from tests.mocks import MockedStep, MockedWorld
from mock import patch


class TestStepCases(TestCase):

    def test_i_have_name_section_configured_type_not_found(self):
        step = MockedStep()
        with self.assertRaises(AssertionError) as err:
            i_have_name_section_configured(step, 'some_name', 'some_invalid_type', 'some_world')

        self.assertEqual(str(err.exception), 'some_invalid_type configuration type does not exist or not implemented '
                                             'yet. Use resource(s), provider(s), variable(s), output(s) or data(s) '
                                             'instead.')

    def test_i_have_name_section_configured_resource_that_supports_tags_found(self):
        step = MockedStep()
        world = MockedWorld()
        name = 'resource that supports tags'
        type_name = 'resource'

        self.assertTrue(i_have_name_section_configured(step, name, type_name, world))
        self.assertEqual(step.context.name, name)
        self.assertEqual(step.context.type, type_name)

    @patch('terraform_compliance.extensions.ext_radish_bdd.world', return_value=MockedWorld())
    def test_i_have_name_section_configured_resource_that_supports_tags_not_found(self, *args):
        step = MockedStep()
        world = MockedWorld()
        del world.config.terraform.resources['provider_type_id']
        name = 'resource that supports tags'
        type_name = 'resource'

        self.assertIsNone(i_have_name_section_configured(step, name, type_name, world))
        self.assertEqual(step.state, 'skipped')

    @patch('terraform_compliance.extensions.ext_radish_bdd.world', return_value=MockedWorld())
    def test_i_have_name_section_configured_resource_not_found(self, *args):
        step = MockedStep()
        world = MockedWorld()
        name = 'non_existent_resource_id'
        type_name = 'resource'

        self.assertIsNone(i_have_name_section_configured(step, name, type_name, world))
        self.assertEqual(step.state, 'skipped')

    def test_i_have_name_section_configured_resource_found(self):
        step = MockedStep()
        world = MockedWorld()
        name = 'resource_type_without_tags'
        type_name = 'resource'

        self.assertTrue(i_have_name_section_configured(step, name, type_name, world))

    @patch('terraform_compliance.extensions.ext_radish_bdd.world', return_value=MockedWorld())
    def test_i_have_name_section_configured_variable_not_found(self, *args):
        step = MockedStep()
        world = MockedWorld()
        name = 'non_existent_variable'
        type_name = 'variable'

        self.assertIsNone(i_have_name_section_configured(step, name, type_name, world))
        self.assertEqual(step.state, 'skipped')

    def test_i_have_name_section_configured_variable_found(self):
        step = MockedStep()
        world = MockedWorld()
        name = 'some_variable'
        type_name = 'variables'

        self.assertTrue(i_have_name_section_configured(step, name, type_name, world))

    @patch('terraform_compliance.extensions.ext_radish_bdd.world', return_value=MockedWorld())
    def test_i_have_name_section_configured_provider_not_found(self, *args):
        step = MockedStep()
        world = MockedWorld()
        name = 'non_existing_provider'
        type_name = 'provider'

        self.assertIsNone(i_have_name_section_configured(step, name, type_name, world))
        self.assertEqual(step.state, 'skipped')

    def test_i_have_name_section_configured_provider_found(self):
        step = MockedStep()
        world = MockedWorld()
        name = 'some_provider'
        type_name = 'providers'

        self.assertTrue(i_have_name_section_configured(step, name, type_name, world))

    @patch('terraform_compliance.extensions.ext_radish_bdd.world', return_value=MockedWorld())
    @patch('terraform_compliance.common.error_handling.world', side_effect=MockedWorld())
    def test_it_condition_contain_something_resource_is_not_dict_failure(self, *args):
        step = MockedStep()
        step.context_sensitive_sentence = 'it must contain something'
        step.context.type = 'resource'
        step.context.stash = ['some_resource']
        with self.assertRaises(Failure):
            self.assertIsNone(it_condition_contain_something(step, 'something'))

    @patch('terraform_compliance.extensions.ext_radish_bdd.world', return_value=MockedWorld())
    @patch('terraform_compliance.common.error_handling.world', side_effect=MockedWorld())
    def test_it_condition_contain_something_resource_not_found(self, *args):
        step = MockedStep()
        step.context_sensitive_sentence = 'it must contain something'
        step.context.type = 'resource'
        step.context.stash = [
            {
                'address': 'some_address',
                'type': 'resource',
                'values': {
                    'something': True
                }
            }
        ]
        with self.assertRaises(Failure) as err:
            self.assertIsNone(it_condition_contain_something(step, 'something else'))
        self.assertEqual(str(err.exception), '{} ({}) does not have {} property.'.format('some_address',
                                                                                         'resource',
                                                                                         'something else'))

    @patch('terraform_compliance.extensions.ext_radish_bdd', return_value=None)
    @patch('terraform_compliance.extensions.ext_radish_bdd.world', return_value=MockedWorld())
    @patch('terraform_compliance.common.error_handling.world', side_effect=MockedWorld())
    def test_it_condition_contain_something_resource_not_found_skip_step(self, *args):
        step = MockedStep()
        step.context_sensitive_sentence = 'it contains something'
        step.context.type = 'resource'
        step.context.stash = [
            {
                'address': 'some_address',
                'type': 'resource',
                'values': {
                    'something': True
                }
            }
        ]

        self.assertIsNone(it_condition_contain_something(step, 'something else'))

    @patch('terraform_compliance.extensions.ext_radish_bdd.world', return_value=MockedWorld())
    def test_it_condition_contain_something_resource_found(self, *args):
        step = MockedStep()
        step.context_sensitive_sentence = 'it contains something'
        step.context.type = 'resource'
        step.context.stash = [
            {
                'address': 'some_address',
                'type': 'resource',
                'values': {
                    'something': True
                }
            }
        ]

        self.assertTrue(it_condition_contain_something(step, 'something'))

    @patch('terraform_compliance.extensions.ext_radish_bdd.world', return_value=MockedWorld())
    def test_it_condition_contain_something_resource_value_is_list(self, *args):
        step = MockedStep()
        step.context_sensitive_sentence = 'it contains something'
        step.context.type = 'resource'
        step.context.stash = [
            {
                'address': 'some_address',
                'type': 'resource',
                'values': [
                    {
                        'key': 'something',
                        'value': 'some_value'
                    }
                ]
            }
        ]

        self.assertTrue(it_condition_contain_something(step, 'something'))

    @patch('terraform_compliance.steps.steps.seek_key_in_dict', return_value=None)
    @patch('terraform_compliance.extensions.ext_radish_bdd.world', return_value=MockedWorld())
    @patch('terraform_compliance.common.error_handling.world', side_effect=MockedWorld())
    def test_it_condition_contain_something_provider_not_found(self, *args):
        step = MockedStep()
        step.context.type = 'provider'
        step.context.stash = []

        self.assertIsNone(it_condition_contain_something(step, 'something'))

    @patch('terraform_compliance.steps.steps.seek_key_in_dict', return_value=True)
    @patch('terraform_compliance.extensions.ext_radish_bdd.world', return_value=MockedWorld())
    def test_it_condition_contain_something_provider_found(self, *args):
        step = MockedStep()
        step.context.type = 'provider'
        step.context.stash = [{'name': 'test'}]

        self.assertTrue(it_condition_contain_something(step, 'something'))

    @patch('terraform_compliance.common.error_handling.world', side_effect=MockedWorld())
    def test_property_is_enabled_not_implemented(self, *args):
        step = MockedStep()
        step.context.stash = [
            {
                'type': 'something'
            }
        ]
        with self.assertRaises(Failure):
            property_is_enabled(step, 'something')

    def test_property_is_enabled_success(self):
        step = MockedStep()
        step.context.stash = [
            {
                'type': 'aws_db_instance',
                'values': {
                    'storage_encrypted': True
                }
            }
        ]
        self.assertTrue(property_is_enabled(step, 'storage_encrypted'))

    @patch('terraform_compliance.common.error_handling.world', side_effect=MockedWorld())
    def test_property_is_enabled_failure(self, *args):
        step = MockedStep()
        step.context.stash = [
            {
                'type': 'aws_db_instance',
                'address': 'resource',
                'values': {
                    'storage_encrypted': False
                }
            }
        ]
        with self.assertRaises(Failure):
            property_is_enabled(step, 'storage_encrypted')

    def test_it_condition_have_proto_protocol_and_port_port_for_cidr_ports_must_only_success(self):
        step = MockedStep()
        step.context.stash = []
        condition = 'not important'
        proto = 'not important'
        port = '1-2'
        cidr = 'not important'
        self.assertTrue(it_condition_have_proto_protocol_and_port_port_for_cidr(step, condition, proto, port, cidr))

    def test_it_condition_have_proto_protocol_and_port_port_for_cidr_ports_with_comma_must_only_success(self):
        step = MockedStep()
        step.context.stash = []
        condition = 'not important'
        proto = 'not important'
        port = '1,2'
        cidr = 'not important'
        self.assertTrue(it_condition_have_proto_protocol_and_port_port_for_cidr(step, condition, proto, port, cidr))

    def test_it_condition_have_proto_protocol_and_port_port_for_cidr_port_success(self):
        step = MockedStep()
        step.context.stash = []
        condition = 'not important'
        proto = 'not important'
        port = '1'
        cidr = 'not important'
        self.assertTrue(it_condition_have_proto_protocol_and_port_port_for_cidr(step, condition, proto, port, cidr))

    def test_i_action_them_count_list_of_list(self):
        step = MockedStep()
        step.context.stash = [1,2,3]
        i_action_them(step, 'count')
        self.assertEqual(step.context.stash, {'values': 3})

    def test_i_action_them_count_list_of_dict_with_resources(self):
        step = MockedStep()
        step.context.stash = [
            {
                'values': {
                    'first': True
                }
            },
            {
                'values': {
                    'second': True
                }
            },
            {
                'values': {
                    'third': True
                }
            }
        ]
        i_action_them(step, 'count')
        self.assertEqual(step.context.stash, {'values': 3})

    def test_i_action_them_count_list_of_dict_with_properties(self):
        step = MockedStep()
        step.context.stash = [{},{},{}]
        step.context.property_name = 'resource'
        i_action_them(step, 'count')
        self.assertEqual(step.context.stash, {'values': 3})



    def test_i_action_them_sum(self):
        step = MockedStep()
        with self.assertRaises(TerraformComplianceNotImplemented):
            i_action_them(step, 'something_else')

    @patch('terraform_compliance.common.error_handling.world', side_effect=MockedWorld())
    def test_i_expect_the_result_is_operator_than_number_more(self, *args):
        step = MockedStep()
        step.context.stash = {'values': 42}
        self.assertIsNone(i_expect_the_result_is_operator_than_number(step, 'more', 41))
        with self.assertRaises(Failure) as err:
            i_expect_the_result_is_operator_than_number(step, 'more', 43)
        self.assertTrue('42 is not more than 43' in str(err.exception))

    @patch('terraform_compliance.common.error_handling.world', side_effect=MockedWorld())
    def test_i_expect_the_result_is_operator_than_number_more_and_equal(self, *args):
        step = MockedStep()
        step.context.stash = {'values': 42}
        self.assertIsNone(i_expect_the_result_is_operator_than_number(step, 'more and equal', 42))
        self.assertIsNone(i_expect_the_result_is_operator_than_number(step, 'more and equal', 41))
        with self.assertRaises(Failure) as err:
            i_expect_the_result_is_operator_than_number(step, 'more and equal', 43)
        self.assertTrue('42 is not more and equal than 43' in str(err.exception))

    @patch('terraform_compliance.common.error_handling.world', side_effect=MockedWorld())
    def test_i_expect_the_result_is_operator_than_number_less(self, *args):
        step = MockedStep()
        step.context.stash = {'values': 42}
        self.assertIsNone(i_expect_the_result_is_operator_than_number(step, 'less', 43))
        with self.assertRaises(Failure) as err:
            i_expect_the_result_is_operator_than_number(step, 'less', 41)
        self.assertTrue('42 is not less than 41' in str(err.exception))

    @patch('terraform_compliance.common.error_handling.world', side_effect=MockedWorld())
    def test_i_expect_the_result_is_operator_than_number_less_and_equal(self, *args):
        step = MockedStep()
        step.context.stash = {'values': 42}
        self.assertIsNone(i_expect_the_result_is_operator_than_number(step, 'less and equal', 43))
        self.assertIsNone(i_expect_the_result_is_operator_than_number(step, 'less and equal', 42))
        with self.assertRaises(Failure) as err:
            i_expect_the_result_is_operator_than_number(step, 'less and equal', 41)
        self.assertTrue('42 is not less and equal than 41' in str(err.exception))

    @patch('terraform_compliance.common.error_handling.world', side_effect=MockedWorld())
    def test_i_expect_the_result_is_operator_than_number_equal(self, *args):
        step = MockedStep()
        step.context.stash = {'values': 42}
        self.assertIsNone(i_expect_the_result_is_operator_than_number(step, 'equal', 42))
        with self.assertRaises(Failure) as err:
            i_expect_the_result_is_operator_than_number(step, 'equal', 41)
        self.assertTrue('42 is not equal to 41' in str(err.exception))

    def test_i_expect_the_result_is_more_than_number_failure(self):
        step = MockedStep()
        step.context.stash = dict(values=3)
        with self.assertRaises(TerraformComplianceNotImplemented) as err:
            self.assertIsNone(i_expect_the_result_is_operator_than_number(step, 'invalid operator', 1))
        self.assertTrue('Invalid operator: invalid operator' in str(err.exception))

    @patch('terraform_compliance.common.error_handling.world', side_effect=MockedWorld())
    def test_it_fails(self, *args):
        step = MockedStep()
        step.context.type = 'some_type'
        step.context.name = 'some_name'
        with self.assertRaises(Failure):
            it_fails(step)

    @patch('terraform_compliance.extensions.ext_radish_bdd.world', return_value=MockedWorld())
    def test_its_key_is_value_not_existent(self, *args):
        step = MockedStep()
        step.context.stash = [
            {
                'type': 'aws_db_instance',
                'some_key': 'some_value',
                'values': {
                    'storage_encrypted': True
                }
            },
            {
                'type': 'aws_db_instance',
                'some_key': 'some_other_value',
                'values': {
                    'storage_encrypted': False
                }
            }
        ]
        its_key_is_value(step, 'something_else', 'some_value')
        self.assertEqual(step.state, 'skipped')

    def test_its_key_is_value_success(self):
        step = MockedStep()
        step.context.stash = [
            {
                'type': 'aws_db_instance',
                'some_key': 'some_value',
                'values': {
                    'storage_encrypted': True
                }
            },
            {
                'type': 'aws_db_instance',
                'some_key': 'some_other_value',
                'values': {
                    'storage_encrypted': False
                }
            }
        ]
        its_key_is_value(step, 'some_key', 'some_value')
        self.assertTrue(type(step.context.stash) is list)
        self.assertEqual(step.context.stash[0]['some_key'], 'some_value')

    def test_its_key_is_value_exist_as_a_list(self):
        step = MockedStep()
        step.context.stash = [
            {
                'type': 'aws_db_instance',
                'some_key': 'some_value[0]',
                'values': {
                    'storage_encrypted': True
                }
            },
            {
                'type': 'aws_db_instance',
                'some_key': 'some_other_value',
                'values': {
                    'storage_encrypted': False
                }
            }
        ]
        its_key_is_value(step, 'some_key', 'some_value')
        self.assertTrue(type(step.context.stash) is list)
        self.assertEqual(step.context.stash[0]['some_key'], 'some_value[0]')

    def test_its_key_is_value_exist_in_values_int(self):
        step = MockedStep()
        step.context.stash = [
            {
                'type': 'aws_db_instance',
                'some_key': 'some_value[0]',
                'values': {
                    'storage_encrypted': 1
                }
            },
            {
                'type': 'aws_db_instance',
                'some_key': 'some_other_value',
                'values': {
                    'storage_encrypted': 2
                }
            }
        ]
        its_key_is_value(step, 'storage_encrypted', 1)
        self.assertTrue(type(step.context.stash) is list)
        self.assertEqual(step.context.stash[0]['some_key'], 'some_value[0]')

    def test_its_key_is_value_exist_in_values_bool(self):
        step = MockedStep()
        step.context.stash = [
            {
                'type': 'aws_db_instance',
                'some_key': 'some_value[0]',
                'values': {
                    'storage_encrypted': True
                }
            },
            {
                'type': 'aws_db_instance',
                'some_key': 'some_other_value',
                'values': {
                    'storage_encrypted': False
                }
            }
        ]
        its_key_is_value(step, 'storage_encrypted', True)
        self.assertTrue(type(step.context.stash) is list)
        self.assertEqual(step.context.stash[0]['some_key'], 'some_value[0]')

    @patch('terraform_compliance.extensions.ext_radish_bdd.world', return_value=MockedWorld())
    def test_find_keys_that_has_kv_structure(self, *args):
        step = MockedStep()
        step.context.stash = [
            {
                'address': 'some_resource.id',
                'type': 'some_resource_type',
                'name': 'some_name',
                'values': [
                    {
                        'key': 'some_other_key',
                        'value': 'some_other_value'
                    },
                    {
                        'key': 'some_key',
                        'value': 'some_value'
                    }
                ]
            }
        ]
        step.context.type = 'resource'
        step.context.name = 'some_name'
        step.context.property_name = 'tags'
        step.context_sensitive_sentence = 'must'
        it_condition_contain_something(step, 'some_key')
        self.assertEqual(step.context.stash[0]['values'], 'some_value')

    @patch('terraform_compliance.common.error_handling.world', side_effect=MockedWorld())
    def test_its_value_condition_match_the_search_regex_regex_null_value_is_parsed(self, *args):
        step = MockedStep()
        step.context.stash = [
            {
                'address': 'some_resource.id',
                'type': 'some_resource_type',
                'name': 'some_name',
                'values': None
            }
        ]
        step.context.type = 'resource'
        step.context.name = 'some_name'
        step.context.property_name = 'tags'
        step.context_sensitive_sentence = 'must'

        with self.assertRaises(Failure):
            its_value_condition_match_the_search_regex_regex(step, 'must', 'something')
        with self.assertRaises(Failure):
            its_value_condition_equal(step, 'must', 'something')

    @patch('terraform_compliance.common.error_handling.world', side_effect=MockedWorld())
    def test_its_value_condition_match_the_search_regex_regex_success(self, *args):
        step = MockedStep()
        step.context.stash = [
            {
                'address': 'some_resource.id',
                'type': 'some_resource_type',
                'name': 'some_name',
                'values': [
                    {
                        'key': 'some_other_key',
                        'value': 'some_other_value'
                    },
                    {
                        'key': 'some_key',
                        'value': 'some_value'
                    }
                ]
            }
        ]
        step.context.type = 'resource'
        step.context.name = 'some_name'
        step.context.property_name = 'tags'
        step.context_sensitive_sentence = 'must'

        self.assertEqual(its_value_condition_match_the_search_regex_regex(step, 'must', 'some_.*'), None)

        with self.assertRaises(Failure):
            self.assertEqual(its_value_condition_match_the_search_regex_regex(step, 'must not', 'some_.*'), None)
            self.assertEqual(its_value_condition_match_the_search_regex_regex(step, 'must not', 'some_other.*'), None)

    @patch('terraform_compliance.common.error_handling.world', side_effect=MockedWorld())
    def test_its_value_condition_equals(self, *args):
        step = MockedStep()
        expected_value = r"https://www.stackoverflow.com[as](.*)\s\t+$"
        step.context.stash = [
            {
                'address': 'some_resource.id',
                'type': 'some_resource_type',
                'name': 'some_name',
                'values': {
                    'some_key': expected_value
                }
            }
        ]
        step.context.type = 'resource'
        step.context.name = 'some_name'
        step.context.property_name = 'tags'
        step.context_sensitive_sentence = 'must'

        self.assertEqual(its_value_condition_equal(step, 'must', expected_value), None)
        self.assertEqual(its_value_condition_equal(step, 'must not', expected_value * 2), None)

        with self.assertRaises(Failure):
            self.assertEqual(its_value_condition_equal(step, 'must', expected_value + ' '), None)
            self.assertEqual(its_value_condition_equal(step, 'must not', expected_value), None)

    def test_its_key_is_not_value_exist_in_values_bool(self):
        step = MockedStep()
        step.context.stash = [
            {
                'type': 'aws_db_instance',
                'some_key': 'some_value[0]',
                'values': {
                    'storage_encrypted': True
                }
            },
            {
                'type': 'aws_db_instance',
                'some_key': 'some_other_value',
                'values': {
                    'storage_encrypted': False
                }
            }
        ]
        its_key_is_not_value(step, 'storage_encrypted', True)
        self.assertTrue(type(step.context.stash) is list)
        self.assertEqual(step.context.stash[0]['some_key'], 'some_other_value')

    @patch('terraform_compliance.extensions.ext_radish_bdd.world', return_value=MockedWorld())
    def test_its_key_is_not_value_not_existent(self, *args):
        step = MockedStep()
        step.context.stash = [
            {
                'type': 'aws_db_instance',
                'some_key': 'some_value',
                'values': {
                    'storage_encrypted': True
                }
            },
            {
                'type': 'aws_db_instance',
                'some_key': 'some_other_value',
                'values': {
                    'storage_encrypted': False
                }
            }
        ]
        its_key_is_not_value(step, 'something_else', 'some_value')
        self.assertEqual(step.state, 'skipped')

    def test_its_key_is_not_value_success(self):
        step = MockedStep()
        step.context.stash = [
            {
                'type': 'aws_db_instance',
                'some_key': 'some_value',
                'values': {
                    'storage_encrypted': True
                }
            },
            {
                'type': 'aws_db_instance',
                'some_key': 'some_other_value',
                'values': {
                    'storage_encrypted': False
                }
            }
        ]
        its_key_is_not_value(step, 'some_key', 'some_value')
        self.assertTrue(type(step.context.stash) is list)
        self.assertEqual(step.context.stash[0]['some_key'], 'some_other_value')

    def test_its_key_is_not_value_exist_as_a_list(self):
        step = MockedStep()
        step.context.stash = [
            {
                'type': 'aws_db_instance',
                'some_key': 'some_value[0]',
                'values': {
                    'storage_encrypted': True
                }
            },
            {
                'type': 'aws_db_instance',
                'some_key': 'some_other_value[0]',
                'values': {
                    'storage_encrypted': False
                }
            }
        ]
        its_key_is_not_value(step, 'some_key', 'some_value')
        self.assertTrue(type(step.context.stash) is list)
        self.assertEqual(step.context.stash[0]['some_key'], 'some_other_value[0]')

    def test_its_key_is_not_value_exist_in_values_int(self):
        step = MockedStep()
        step.context.stash = [
            {
                'type': 'aws_db_instance',
                'some_key': 'some_value[0]',
                'values': {
                    'storage_encrypted': 1
                }
            },
            {
                'type': 'aws_db_instance',
                'some_key': 'some_other_value',
                'values': {
                    'storage_encrypted': 2
                }
            }
        ]
        its_key_is_not_value(step, 'storage_encrypted', 1)
        self.assertTrue(type(step.context.stash) is list)
        self.assertEqual(step.context.stash[0]['some_key'], 'some_other_value')

    @patch('terraform_compliance.common.error_handling.world', side_effect=MockedWorld())
    def test_its_value_condition_contain(self, *args):
        step = MockedStep()
        step.context.stash = [
            {
                'values': ['foo', 'bar']
            }
        ]
        step.context.property_name = 'some_thing'
        self.assertEqual({'values': ['foo', 'bar']},
                         its_value_condition_contain(step, 'must', 'foo'))
        self.assertEqual({'values': ['foo', 'bar']}, its_value_condition_contain(step, 'must not', 'baz'))
        with self.assertRaises(Failure):
            its_value_condition_contain(step, 'must', 'baz')
        with self.assertRaises(Failure):
            its_value_condition_contain(step, 'must not', 'foo')



property_match_list = {
    'encryption at rest': {
        'aws_db_instance': 'storage_encrypted',
        'ebs_block_device': 'encrypted',
        'aws_ebs_volume': 'encrypted',
        'azurerm_storage_account': 'enable_blob_encryption',
        'azurerm_sql_database': 'encryption',
        'aws_s3_bucket': 'server_side_encryption_configuration',
        'aws_elasticache_replication_group': 'at_rest_encryption_enabled',
        'aws_emr_security_configuration': 'EnableAtRestEncryption'
    },
    'encryption in flight': {
        'aws_elasticache_replication_group': 'transit_encryption_enabled',
        'aws_emr_security_configuration': 'EnableInTransitEncryption'
    }
}

resource_name = {
    'AWS Auto-Scaling Group': 'aws_autoscaling_group',
    'AWS EBS volume': 'aws_ebs_volume',
    'AWS EC2 instance': 'aws_instance',
    'AWS EFS file system': 'aws_efs_file_system',
    'AWS ELB resource': 'aws_elb',
    'AWS EMR Security Configuration': 'aws_emr_security_configuration',
    'AWS Internet Gateway': 'aws_internet_gateway',
    'AWS Launch Configuration': 'aws_launch_configuration',
    'AWS RDS instance': 'aws_db_instance',
    'AWS Route Table': 'aws_route_table',
    'AWS S3 Bucket': 'aws_s3_bucket',
    'AWS Security Group': 'aws_security_group',
    'AWS VPC': 'aws_vpc',
    'Azure SQL Database': 'azurerm_sql_database',
    'Azure Storage Account': 'azurerm_storage_account',
    'AWS Subnet': 'aws_subnet',
    'AWS ElastiCache Replication Group': 'aws_elasticache_replication_group',
    'AWS Elastic Cache Replication Group': 'aws_elasticache_replication_group'
}



# -*- coding: utf-8 -*-

from radish import world, given, when, then, step
from terraform_compliance.steps import property_match_list
from terraform_compliance.common.helper import convert_resource_type, find_root_by_key, seek_key_in_dict
from terraform_compliance.common.helper import search_regex_in_list, seek_regex_key_in_dict_values, jsonify, Null, EmptyStash
from terraform_compliance.common.helper import get_resource_name_from_stash, get_resource_address_list_from_stash
from terraform_compliance.common.helper import remove_mounted_resources, search_regex_in_list, seek_value_in_dict
from terraform_compliance.extensions.security_groups import SecurityGroup
from terraform_compliance.extensions.ext_radish_bdd import skip_step
from terraform_compliance.extensions.ext_radish_bdd import custom_type_any
from terraform_compliance.extensions.ext_radish_bdd import custom_type_condition
from terraform_compliance.extensions.ext_radish_bdd import custom_type_section
from terraform_compliance.extensions.ext_radish_bdd import custom_type_prop
import re
from terraform_compliance.common.exceptions import Failure, TerraformComplianceNotImplemented
from terraform_compliance.common.exceptions import TerraformComplianceInternalFailure
from terraform_compliance.common.error_handling import Error
from terraform_compliance.main import Step
from terraform_compliance.common.defaults import Defaults
from terraform_compliance.common.bdd_tags import look_for_bdd_tags

# TODO: Figure out how the IAM policies/statements shown in the plan.out
# TODO: Implement an IAM Compliance via https://github.com/Netflix-Skunkworks/policyuniverse

types_list = ['resource', 'variable', 'provider', 'data', 'resource that supports tags']


@given(u'I have {name:ANY} defined')
@given(u'I have {name:ANY} {type_name:SECTION} configured')
def i_have_name_section_configured(_step_obj, name, type_name='resource', _terraform_config=world):
    '''
    Finds given resource or variable by name and returns it. Skips the step (and further steps) if it is not found.

    :param _step_obj: Internal, step object for radish.
    :param name: String of the name of the resource_type or variable.
    :param type_name: String of the type, either resource(s) or variable(s)
    :param _terraform_config: Internal, terraform configuration.
    :return:
    '''
    assert (type_name in ['resource', 'resources',
                          'variable', 'variables',
                          'output', 'outputs',
                          'provider', 'providers',
                          'data', 'datas']), \
        '{} configuration type does not exist or not implemented yet. ' \
        'Use resource(s), provider(s), variable(s), output(s) or data(s) instead.'.format(type_name)

    if type_name.endswith('s'):
        type_name = type_name[:-1]

    # Process the tags
    _step_obj = look_for_bdd_tags(_step_obj)

    if name in ('a resource', 'any resource', 'resources'):
        _step_obj.context.type = type_name
        _step_obj.context.name = name
        _step_obj.context.stash = [obj for key, obj in _terraform_config.config.terraform.resources_raw.items()]
        _step_obj.context.addresses = get_resource_address_list_from_stash(_step_obj.context.stash)
        _step_obj.context.property_name = type_name
        return True

    elif name in ('an output', 'any output', 'outputs'):
        _step_obj.context.type = 'output'
        _step_obj.context.name = name
        _step_obj.context.stash = [obj for key, obj in _terraform_config.config.terraform.configuration['outputs'].items()]
        _step_obj.context.addresses = get_resource_address_list_from_stash(_terraform_config.config.terraform.configuration['outputs'])
        _step_obj.context.property_name = 'output'
        return True

    elif name in ('a variable', 'any variable', 'variables'):
        _step_obj.context.type = 'variable'
        _step_obj.context.name = name
        _step_obj.context.stash = [obj for key, obj in _terraform_config.config.terraform.configuration['variables'].items()]
        _step_obj.context.addresses = 'variable'
        _step_obj.context.property_name = 'variable'
        return True

    elif name == 'resource that supports tags':
        resource_types_supports_tags = find_root_by_key(_terraform_config.config.terraform.resources_raw,
                                                        'tags',
                                                        return_key='type')
        resource_list = []
        for resource_type in resource_types_supports_tags:
            # Issue-168: Mounted resources causes problem on recursive searching for resources that supports tags
            #            We are removing all mounted resources here for future steps, since we don't need them for
            #            tags checking.
            found_resources = remove_mounted_resources(_terraform_config.config.terraform.find_resources_by_type(resource_type))
            resource_list.extend(found_resources)

        if resource_list:
            _step_obj.context.type = type_name
            _step_obj.context.name = name
            _step_obj.context.stash = resource_list
            _step_obj.context.addresses = get_resource_address_list_from_stash(resource_list)
            _step_obj.context.property_name = type_name
            return True

    elif type_name == 'resource':
        name = convert_resource_type(name)
        resource_list = _terraform_config.config.terraform.find_resources_by_type(name)

        if resource_list:
            _step_obj.context.type = type_name
            _step_obj.context.name = name
            _step_obj.context.stash = resource_list
            _step_obj.context.addresses = get_resource_address_list_from_stash(resource_list)
            _step_obj.context.property_name = type_name
            return True

    elif type_name == 'variable':
        found_variable = _terraform_config.config.terraform.variables.get(name, None)

        if found_variable:
            _step_obj.context.type = type_name
            _step_obj.context.name = name
            _step_obj.context.stash = found_variable
            _step_obj.context.addresses = name
            _step_obj.context.property_name = type_name
            return True


    elif type_name == 'output':
        found_output = _terraform_config.config.terraform.outputs.get(name, None)

        if found_output:
            _step_obj.context.type = type_name
            _step_obj.context.name = name
            _step_obj.context.stash = found_output
            _step_obj.context.addresses = name
            _step_obj.context.property_name = type_name
            return True

    elif type_name == 'provider':
        found_provider = _terraform_config.config.terraform.get_providers_from_configuration(name)

        if found_provider:
            _step_obj.context.type = type_name
            _step_obj.context.name = name
            _step_obj.context.stash = found_provider
            _step_obj.context.addresses = name
            _step_obj.context.address = name
            _step_obj.context.property_name = type_name
            return True

    elif type_name == 'data':
        name = convert_resource_type(name)
        data_list = _terraform_config.config.terraform.find_data_by_type(name)

        if data_list:
            _step_obj.context.type = type_name
            _step_obj.context.name = name
            _step_obj.context.stash = data_list
            _step_obj.context.addresses = name
            _step_obj.context.address = name
            _step_obj.context.property_name = type_name
            return True

    skip_step(_step_obj, name)


@when(u'its {key:PROPERTY} is {value:PROPERTY}')
@when(u'its {key:PROPERTY} has {value:PROPERTY}')
@when(u'its {key:PROPERTY} includes {value:PROPERTY}')
@when(u'its {key:PROPERTY} contains {value:PROPERTY}')
def its_key_is_value(_step_obj, key, value):
    orig_key = key
    if key == 'reference':
        key = Defaults.address_pointer

    key = str(key).lower()
    found_list = []
    for obj in _step_obj.context.stash:
        object_key = obj.get(key, Null)

        if object_key is Null:
            object_key = obj.get('values', {})
            if isinstance(object_key, list):
                object_keys = []
                for object_key_element in object_key:
                    if isinstance(object_key_element, dict):
                        filtered_key = object_key_element.get(key)
                        if isinstance(filtered_key, str) and filtered_key.lower() == value.lower():
                            found_list.append(object_key_element)
                    else:
                        object_keys.append(object_key_element.get(key, Null))

                object_key = [keys for keys in object_keys if keys is not Null]
            else:
                object_key = object_key.get(key, Null)

        if isinstance(object_key, str):
            if "[" in object_key:
                object_key = object_key.split('[')[0]

            if object_key.lower() == value.lower():
                found_list.append(obj)

        elif isinstance(object_key, (int, bool)) and object_key == value:
            found_list.append(obj)

        elif isinstance(object_key, list) and value in object_key:
            found_list.append(obj)

        elif isinstance(object_key, dict) and (value in object_key.keys()):
            found_list.append(obj)

    if found_list != []:
        _step_obj.context.stash = found_list
        _step_obj.context.addresses = get_resource_address_list_from_stash(found_list)
    else:
        skip_step(_step_obj, message='Can not find {} {} in {}.'.format(value, orig_key,
                                                                        ', '.join(_step_obj.context.addresses)))


@when(u'its {key:PROPERTY} is not {value:PROPERTY}')
@when(u'its {key:PROPERTY} has not {value:PROPERTY}')
@when(u'its {key:PROPERTY} does not include {value:PROPERTY}')
@when(u'its {key:PROPERTY} does not contain {value:PROPERTY}')
def its_key_is_not_value(_step_obj, key, value):
    orig_key = key
    if key == 'reference':
        key = Defaults.address_pointer

    key = str(key).lower()
    found_list = []
    for obj in _step_obj.context.stash:
        object_key = obj.get(key, Null)

        if object_key is Null:
            object_key = obj.get('values', {})
            if isinstance(object_key, list):
                object_keys = []
                for object_key_element in object_key:
                    if object_key_element.get(key, Null) != value:
                        object_keys.append(object_key_element.get(key, Null))

                object_key = [keys for keys in object_keys if keys is not Null]
            else:
                if object_key.get(key, Null) != value:
                    object_key = object_key.get(key, Null)

        if isinstance(object_key, str):
            if "[" in object_key:
                object_key = object_key.split('[')[0]

            if object_key != value:
                found_list.append(obj)

        elif isinstance(object_key, (int, bool)) and object_key != value:
            found_list.append(obj)

        elif isinstance(object_key, list) and value not in object_key:
            found_list.append(obj)

        elif isinstance(object_key, dict) and (value in object_key.keys()):
            found_list.append(obj)

    if found_list != []:
        _step_obj.context.stash = found_list
        _step_obj.context.addresses = get_resource_address_list_from_stash(found_list)
    else:
        skip_step(_step_obj, message='Can not find {} {} in {}.'.format(value, orig_key,
                                                                        ', '.join(_step_obj.context.addresses)))


@when(u'it contain {something:ANY}')
@when(u'they have {something:ANY}')
@when(u'it has {something:ANY}')
@when(u'it contains {something:ANY}')
@then(u'it must contain {something:ANY}')
def it_condition_contain_something(_step_obj, something):
    prop_list = []

    if _step_obj.context.type in ('resource', 'data'):
        for resource in _step_obj.context.stash:
            if not isinstance(resource, dict) \
                    or 'values' not in resource \
                    or 'address' not in resource \
                    or 'type' not in resource:
                resource = {'values': resource,
                            'address': resource,
                            'type': _step_obj.context.name}

            values = resource.get('values', resource.get('expressions', {}))
            if not values:
                values = seek_key_in_dict(resource, something)

            found_value = Null
            found_key = Null
            if isinstance(values, dict):
                found_key = values.get(something, seek_key_in_dict(values, something))
                if not isinstance(found_key, list):
                    found_key = [{something: found_key}]

                if len(found_key):
                    found_key = found_key[0] if len(found_key) == 1 and something in found_key[0] else found_key

                    if isinstance(found_key, dict):
                        found_value = jsonify(found_key.get(something, found_key))
                    else:
                        found_value = found_key
            elif isinstance(values, list):
                found_value = []

                for value in values:

                    if isinstance(value, dict):
                        # First search in the keys
                        found_key = seek_key_in_dict(value, something)

                        # Then search in the values with 'key'
                        if not found_key:
                            found_key = seek_regex_key_in_dict_values(value, 'key', something)

                            if found_key:
                                found_key = found_key[0]
                                found_value = value.get('value')
                                break

                    if found_key is not Null and len(found_key):
                        found_key = found_key[0] if len(found_key) == 1 else found_key

                        if isinstance(found_key, dict):
                            found_value.append(jsonify(found_key.get(something, found_key)))

            if isinstance(found_value, dict) and 'constant_value' in found_value:
                found_value = found_value['constant_value']

            if found_value is not Null and found_value != [] and found_value != '' and found_value != {}:
                prop_list.append({'address': resource['address'],
                                  'values': found_value,
                                  'type': _step_obj.context.name})

            elif 'must' in _step_obj.context_sensitive_sentence:
                Error(_step_obj, '{} ({}) does not have {} property.'.format(resource['address'],
                                                                             resource.get('type', ''),
                                                                             something))

        if prop_list:
            _step_obj.context.stash = prop_list
            _step_obj.context.property_name = something
            return True

        if _step_obj.state != Step.State.FAILED:
            skip_step(_step_obj,
                      resource=_step_obj.context.name,
                      message='Can not find any {} property for {} resource in '
                              'terraform plan.'.format(something, _step_obj.context.name))

    elif _step_obj.context.type == 'provider':
        for provider_data in _step_obj.context.stash:
            values = seek_key_in_dict(provider_data, something)

            if values:
                _step_obj.context.stash = values
                _step_obj.context.property_name = something
                _step_obj.context.address = '{}.{}'.format(provider_data.get('name', _step_obj.context.addresses),
                                                           provider_data.get('alias', "\b"))
                return True
            elif 'must' in _step_obj.context_sensitive_sentence:
                Error(_step_obj, '{} {} does not have {} property.'.format(_step_obj.context.addresses,
                                                                           _step_obj.context.type,
                                                                           something))
        if 'must' in _step_obj.context_sensitive_sentence:
            Error(_step_obj, '{} {} does not have {} property.'.format(_step_obj.context.addresses,
                                                                       _step_obj.context.type,
                                                                       something))
    if _step_obj.state != Step.State.FAILED:
        skip_step(_step_obj,
                  resource=_step_obj.context.name,
                  message='Skipping the step since {} type does not have {} property.'.format(_step_obj.context.type,
                                                                                              something))


@then(u'{something:ANY} is be enabled')
@then(u'{something:ANY} must be enabled')
def property_is_enabled(_step_obj, something):
    for resource in _step_obj.context.stash:
        if isinstance(resource, dict):
            if something in property_match_list:
                something = property_match_list[something].get(resource['type'], something)

            property_value = seek_key_in_dict(resource.get('values', {}), something)

            if len(property_value):
                property_value = property_value[0]

                if isinstance(property_value, dict):
                    property_value = property_value.get(something, Null)

            if not property_value:
                Error(_step_obj, 'Resource {} does not have {} property enabled '
                                 '({}={}).'.format(resource.get('address', "resource"),
                                                   something,
                                                   something,
                                                   property_value))
    return True


@then(u'it {condition:ANY} have {proto:ANY} protocol and port {port} for {cidr:ANY}')
def it_condition_have_proto_protocol_and_port_port_for_cidr(_step_obj, condition, proto, port, cidr):
    searching_for=dict(port=port, protocol=proto, cidr_blocks=cidr)

    for sg in _step_obj.context.stash:
        if sg['type'] != 'aws_security_group':
            raise TerraformComplianceInternalFailure('This method can only be used for aws_security_group resources '
                                                     'for now. You tried to used it on {}'.format(sg['type']))

        sg_obj = SecurityGroup(searching_for, sg['values'], address=sg['address'])
        if condition == 'must only':
            sg_obj.must_only_have()
        elif condition == 'must':
            sg_obj.must_have()
        elif condition == 'must not':
            sg_obj.must_not_have()
        else:
            raise TerraformComplianceInternalFailure('You can only use "must have", "must not have" and "must only have"'
                                                     'conditions on this step for now.'
                                                     'You tried to use "{}"'.format(condition))
        result, message = sg_obj.validate()

        if result is False:
            Error(_step_obj, message)

    return True


@when(u'I {action_type:ANY} it')
@when(u'I {action_type:ANY} them')
@when(u'I {action_type:ANY} the value')
def i_action_them(_step_obj, action_type):
    if action_type == "count":
        # WARNING: Only case where we set stash as a dictionary, instead of a list.
        if isinstance(_step_obj.context.stash, list):

            # This means we are directly started counting without drilling down any property
            # Thus, our target for the count is stash itself.
            if _step_obj.context.property_name in types_list:
                _step_obj.context.stash = dict(values=len(_step_obj.context.stash))

            else:
                if isinstance(_step_obj.context.stash[0], dict):
                    if _step_obj.context.stash[0].get('values'):
                        _step_obj.context.stash = seek_key_in_dict(_step_obj.context.stash, 'values')
                        count = 0
                        for result in _step_obj.context.stash:
                            count += len(result.get('values', {})) if result.get('values') else 1

                        _step_obj.context.stash = dict(values=count)

                else:
                    _step_obj.context.stash = dict(values=len(_step_obj.context.stash))
    else:
        raise TerraformComplianceNotImplemented('Invalid action_type in the scenario: {}'.format(action_type))


@then(u'its value must be {operator:ANY} than {number:d}')
@then(u'I expect the result is {operator:ANY} than {number:d}')
@then(u'its value must be {operator:ANY} to {number:d}')
@then(u'I expect the result is {operator:ANY} to {number:d}')
def i_expect_the_result_is_operator_than_number(_step_obj, operator, number, _stash=EmptyStash):

    def fail(assertion, message):
        try:
            assert assertion, 'Failed'
        except AssertionError as e:
            Error(_step_obj, 'for {} on {}. {}.'.format(_step_obj.context.address,
                                                        _step_obj.context.property_name,
                                                        message))

    values = _step_obj.context.stash if _stash is EmptyStash else _stash

    if isinstance(values, list):
        for value_set in values:
            i_expect_the_result_is_operator_than_number(_step_obj, operator, number, _stash=value_set)

    elif isinstance(values, dict):
        _step_obj.context.property_name = values.get('type', _step_obj.context.property_name)
        _step_obj.context.address = values.get('address', _step_obj.context.addresses)

        if isinstance(_step_obj.context.address, list) and len(_step_obj.context.address) == 1:
            _step_obj.context.address = _step_obj.context.address[0]

        i_expect_the_result_is_operator_than_number(_step_obj, operator, number, values.get('values', Null))

    elif isinstance(values, (int, str)):
        values = int(values)
        if operator in ('more', 'greater', 'bigger'):
            fail(values > number, '{} is not more than {}'.format(values, number))
        elif operator in ('more and equal', 'greater and equal', 'bigger and equal'):
            fail(values >= number, '{} is not more and equal than {}'.format(values, number))
        elif operator in ('less', 'lesser', 'smaller'):
            fail(values < number, '{} is not less than {}'.format(values, number))
        elif operator in ('less and equal', 'lesser and equal', 'smaller and equal'):
            fail(values <= number, '{} is not less and equal than {}'.format(values, number))
        elif operator in ('equal',):
            fail(values == number, '{} is not equal to {}'.format(values, number))
        else:
            raise TerraformComplianceNotImplemented('Invalid operator: {}'.format(operator))

    elif isinstance(values, Null):
        raise TerraformComplianceNotImplemented('Null/Empty value found on {}'.format(_step_obj.context.type))


@then(u'its value {condition:ANY} match the "{search_regex}" regex')
def its_value_condition_match_the_search_regex_regex(_step_obj, condition, search_regex, _stash=EmptyStash):
    def fail(condition, name=None):
        text = 'matches' if condition == 'must not' else 'does not match'
        name = name if (name is not None or name is not False) else _step_obj.context.name
        pattern = 'Null/None' if regex == '\x00' else regex
        Error(_step_obj, '{} property in {} {} {} with {} regex. '
                         'It is set to {}.'.format(_step_obj.context.property_name,
                                                   name,
                                                   _step_obj.context.type,
                                                   text,
                                                   pattern,
                                                   values))

    regex = r'{}'.format(search_regex)
    values = _step_obj.context.stash if _stash is EmptyStash else _stash

    if isinstance(values, (str, int, bool)) or values is None:
        matches = re.match(regex, str(values), flags=re.IGNORECASE)

        if (condition == 'must' and matches is None) or (condition == "must not" and matches is not None):
            _stash = get_resource_name_from_stash(_step_obj.context.stash, _stash, _step_obj.context.address)
            fail(condition, name=_stash.get('address'))

    elif isinstance(values, list):
        for value in values:
            its_value_condition_match_the_search_regex_regex(_step_obj, condition, search_regex, value)

    elif isinstance(values, dict):
        if not hasattr(_step_obj.context, 'address'):
            _step_obj.context.address = None

        _step_obj.context.address = values.get('address', _step_obj.context.address)

        if 'values' in values:
            if values['values'] is None and regex == '\x00' and condition == 'must not':
                values = values['values']
                fail(condition, name=_stash.get('address'))
            else:
                its_value_condition_match_the_search_regex_regex(_step_obj, condition, search_regex, values.get('values'))

        else:
            for key, value in values.items():
                its_value_condition_match_the_search_regex_regex(_step_obj, condition, search_regex, value)


@then(u'its value {condition:ANY} be {match:ANY}')
def its_value_condition_equal(_step_obj, condition, match, _stash=EmptyStash):
    its_value_condition_match_the_search_regex_regex(_step_obj, condition, "^" + re.escape(match) + "$", _stash)


@then(u'its value {condition:ANY} contain {value:ANY}')
def its_value_condition_contain(_step_obj, condition, value, _stash=EmptyStash):
    if condition not in ('must', 'must not'):
        raise TerraformComplianceNotImplemented('Condition should be one of: `must`, `must not`')

    values = _step_obj.context.stash if _stash is EmptyStash else _stash
    # TODO: Update here for checking values in a list or dict.

    if isinstance(values, list):
        for elem in values:
            values = its_value_condition_contain(_step_obj, condition, value, elem)

    if isinstance(values, (int, bool, str, float)):
        values = dict(values=values,
                      address=_step_obj.context.address if hasattr(_step_obj.context, 'address') else _step_obj.context.addresses)

    found_values = seek_value_in_dict(value, values)
    condition = condition == 'must'

    if condition and not found_values:
        if isinstance(values, list):
            objects = []
            for elem in values:
                objects.append(elem.get('address', '???'))
            objects = ', '.join(objects)
        else:
            objects = values.get('address')

        Error(_step_obj, '{} could not found in {}.'.format(value, objects))

    elif not condition and found_values:
        Error(_step_obj, '{} found in {}.'.format(value,
                                                  get_resource_name_from_stash(found_values).get('address')))

    return values


def _its_value_condition_contain(_step_obj, condition, value, values):
    if isinstance(values, list):
        values = [str(v) for v in values]
        _fail_text = 'did not contain' if condition == 'must' else 'contains'
        fail_message = '{} property in {} {} {}, it is set to {}'.format(
            _step_obj.context.property_name,
            _step_obj.context.name,
            _fail_text,
            value,
            values,
        )
        if condition == 'must':
            assert value in values, fail_message
        else:
            assert value not in values, fail_message
    else:
        raise TerraformComplianceInternalFailure('Can only check that if list contains value')


@then(u'the scenario fails')
@then(u'the scenario should fail')
@then(u'the scenario must fail')
@then(u'it fails')
@then(u'it should fail')
@then(u'it must fail')
def it_fails(_step_obj):
    Error(_step_obj, 'Forcefully failing the scenario on {} ({}) {}'.format(_step_obj.context.name,
                                                                            ', '.join(_step_obj.context.addresses),
                                                                            _step_obj.context.type))


@then(u'its value {condition:ANY} be null')
def its_value_condition_be_null(_step_obj, condition):
    its_value_condition_match_the_search_regex_regex(_step_obj, condition, u'\x00')
    its_value_condition_match_the_search_regex_regex(_step_obj, condition, u'^$')
    its_value_condition_match_the_search_regex_regex(_step_obj, condition, u'^null$')


@then(u'it must have "{reference_address}" referenced')
def it_must_have_reference_address_referenced(_step_obj, reference_address):
    if _step_obj.context.stash:
        for resource in _step_obj.context.stash:
            if isinstance(resource, dict):
                if Defaults.address_pointer in resource and search_regex_in_list(reference_address,
                                                                                 resource[Defaults.address_pointer]):
                    return True
            else:
                raise TerraformComplianceInternalFailure('Unexpected resource structure: {}'.format(resource))

            Error(_step_obj, '{} is not referenced within {}.'.format(reference_address, resource.get('address')))
    else:
        Error(_step_obj, 'No entities found for this step to process. Check your filtering steps in this scenario.')


@then('its {key:ANY} {condition:ANY} be {value:ANY}')
@then('its {key:ANY} property {condition:ANY} be {value:ANY}')
@then('its {key:ANY} key {condition:ANY} be {value:ANY}')
def its_key_condition_be_value(_step_obj, key, condition, value, stash=Null, depth=0):
    if condition not in ('must', 'must not'):
        raise TerraformComplianceNotImplemented('This step only accepts "must" and "must not" as a condition.')

    if stash is Null:
        stash = _step_obj.context.stash

    if not stash or stash is Null:
        Error(_step_obj, 'No entities found for this step to process. Check your filtering steps in this scenario.')
        return False

    found_values = []
    for entity in stash:
        if isinstance(entity, dict):
            found_values.extend(seek_regex_key_in_dict_values(entity, key, value))
        elif isinstance(entity, list):
            for element in entity:
                found_values.extend(its_key_condition_be_value(_step_obj, key, condition, element, entity, depth+1))
        elif isinstance(entity ,(str, int, bool)) and (str(entity).lower == key.lower or str(entity) == value.lower):
            found_values.append(entity)

    # Return the values to the parent call.
    if depth > 0:
        return found_values

    condition = condition == 'must'
    found_values = [values for values in found_values if values is not None]

    obj_address = _step_obj.context.name
    if hasattr(_step_obj.context, 'address'):
        obj_address = _step_obj.context.address
    elif hasattr(_step_obj.context, 'addresses'):
        obj_address = ', '.join(_step_obj.context.addresses)

    if found_values and not condition:
        Error(_step_obj, 'Found {}({}) in {} property of {}.'.format(value, ', '.join(found_values), key, obj_address))
    elif not found_values and condition:
        Error(_step_obj, 'Can not find {} in {} property of {}.'.format(value, key, obj_address))

    return True


@then('I flatten all values found')
def i_flatten_everything_found(_step_obj):
    flattened_list = []
    addresses_flattened = []
    for each in _step_obj.context.stash:
        flattened_list.append(each.get('values', each))
        addresses_flattened.append(each.get('address', each))

    if flattened_list:
        _step_obj.context.stash = {
            'address': ', '.join(addresses_flattened),
            'values': flattened_list
        }

    return flattened_list





class TerraformComplianceInvalidConfig(Exception):
    pass


class TerraformComplianceInvalidConfigurationType(Exception):
    pass


class Failure(Exception):
    pass


class TerraformComplianceNotImplemented(Exception):
    pass


class TerraformComplianceInternalFailure(Exception):
    pass


class TerraformComplianceInvalidData(Exception):
    pass
